#include <windows.h>
#include <streams.h>
#include <initguid.h>
#if (1100 > _MSC_VER)
#include <olectlid.h>
#else
#include <olectl.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <tchar.h>

#include "ave_uids.h"
#include "inf_ave.h"
#include "ave_prop.h"
#include "avienc.h"
#include "resource.h"

extern HRESULT initAVI(LPSTR filename, LPBITMAPINFOHEADER lpbi, DWORD dwRate, DWORD dwScale, BOOL showOpts);
extern HRESULT writeAVI(BYTE *data, DWORD len);
extern void closeAVI();
extern LONG	frameNr;

static void multiply(unsigned int x,unsigned int *p);
static void rescale(unsigned char **inpic, int oldw, int oldh,int neww, int newh, unsigned char *outpic);

// Setup information

const AMOVIESETUP_MEDIATYPE sudPinTypes =
{
    &MEDIATYPE_Video,       // Major type
    &MEDIASUBTYPE_NULL      // Minor type
};

const AMOVIESETUP_PIN sudpPins =
{ L"Input",             // Pins string name
  FALSE,                // Is it rendered
  FALSE,                // Is it an output
  FALSE,                // Are we allowed none
  FALSE,                // And allowed many
  &CLSID_NULL,          // Connects to filter
  L"Output",            // Connects to pin
  1,                    // Number of types
  &sudPinTypes          // Pin information
};

const AMOVIESETUP_FILTER sudAVIEncoder =
{
    &CLSID_AVIEncoder,      // Filter CLSID
    L"AVI Encoder",         // String name
    MERIT_DO_NOT_USE,       // Filter merit
    1,                      // Number of pins
    &sudpPins                // Pin information
};

// List of class IDs and creator functions for the class factory. This
// provides the link between the OLE entry point in the DLL and an object
// being created. The class factory will call the static CreateInstance

CFactoryTemplate g_Templates[] = {
    { L"AVIEncoder"
    , &CLSID_AVIEncoder
    , CAVIEncoder::CreateInstance
    , NULL
    , &sudAVIEncoder }
  ,
    { L"AVIEncoder Property Page"
    , &CLSID_AVIEncoderPropertyPage
    , CAVIEncoderProperties::CreateInstance }
};
int g_cTemplates = sizeof(g_Templates) / sizeof(g_Templates[0]);


//
// DllRegisterServer
//
// Handles sample registry and unregistry
//
STDAPI DllRegisterServer()
{
    return AMovieDllRegisterServer2( TRUE );

} // DllRegisterServer


//
// DllUnregisterServer
//
STDAPI DllUnregisterServer()
{
    return AMovieDllRegisterServer2( FALSE );

} // DllUnregisterServer


// Constructor

CAVIEncoderFilter::CAVIEncoderFilter(CAVIEncoder *pAVIEncoder,
                         LPUNKNOWN pUnk,
                         CCritSec *pLock,
                         HRESULT *phr) :
    CBaseFilter(NAME("CAVIEncoderFilter"), pUnk, pLock, CLSID_AVIEncoder),
    m_pAVIEncoder(pAVIEncoder)
{
}


//
// GetPin
//
CBasePin * CAVIEncoderFilter::GetPin(int n)
{
    if (n == 0) {
        return m_pAVIEncoder->m_pPin;
    } else {
        return NULL;
    }
}


//
// GetPinCount
//
int CAVIEncoderFilter::GetPinCount()
{
    return 1;
}


//
// Stop
//
// Overriden to close the AVIEncoder file
//
STDMETHODIMP CAVIEncoderFilter::Stop()
{
    CAutoLock cObjectLock(m_pLock);

    return CBaseFilter::Stop();
}


//
// Pause
//
// Overriden to open the AVIEncoder file
//
STDMETHODIMP CAVIEncoderFilter::Pause()
{
    CAutoLock cObjectLock(m_pLock);

    return CBaseFilter::Pause();
}


//
// Run
//
// Overriden to open the AVIEncoder file
//
STDMETHODIMP CAVIEncoderFilter::Run(REFERENCE_TIME tStart)
{
    CAutoLock cObjectLock(m_pLock);

    return CBaseFilter::Run(tStart);
}


//
//  Definition of CAVIEncoderInputPin
//
CAVIEncoderInputPin::CAVIEncoderInputPin(CAVIEncoder *pAVIEncoder,
                             LPUNKNOWN pUnk,
                             CBaseFilter *pFilter,
                             CCritSec *pLock,
                             CCritSec *pReceiveLock,
                             HRESULT *phr) :

    CRenderedInputPin(NAME("CAVIEncoderInputPin"),
                  pFilter,                   // Filter
                  pLock,                     // Locking
                  phr,                       // Return code
                  L"Input"),                 // Pin name
    m_pReceiveLock(pReceiveLock),
    m_pAVIEncoder(pAVIEncoder),
    m_tLast(0)
{
}


//
// CheckMediaType
//
// Check if the pin can support this specific proposed type and format
//
HRESULT CAVIEncoderInputPin::CheckMediaType(const CMediaType *pMediaType)
{
    if (IsEqualGUID(*pMediaType->Type(), MEDIATYPE_Video)) {
        if (IsEqualGUID(*pMediaType->Subtype(), MEDIASUBTYPE_RGB24)) {
            VIDEOINFOHEADER *pvi = (VIDEOINFOHEADER *) pMediaType->Format();
			{
				if(pvi->bmiHeader.biBitCount == 24)
				{
					return S_OK;
				}
			}
        }
    }

    return S_FALSE;
}


//
// BreakConnect
//
// Break a connection
//
HRESULT CAVIEncoderInputPin::BreakConnect()
{
    if (m_pAVIEncoder->m_pPosition != NULL) {
        m_pAVIEncoder->m_pPosition->ForceRefresh();
    }
    return CRenderedInputPin::BreakConnect();
}


//
// ReceiveCanBlock
//
// We don't hold up source threads on Receive
//
STDMETHODIMP CAVIEncoderInputPin::ReceiveCanBlock()
{
    return S_FALSE;
}


//
// Receive
//
// Do something with this media sample
//
STDMETHODIMP CAVIEncoderInputPin::Receive(IMediaSample *pSample)
{
	CAutoLock lock(m_pReceiveLock);

	if(m_pAVIEncoder->m_aviinit==2)
		return CO_E_FAILEDTOCREATEFILE;

	HRESULT	hr;
	int		crop_width, crop_height;

    VIDEOINFOHEADER *pvi = (VIDEOINFOHEADER *) m_mt.Format();

	crop_width=pvi->bmiHeader.biWidth - m_pAVIEncoder->m_left - m_pAVIEncoder->m_right;
	crop_height=pvi->bmiHeader.biHeight -m_pAVIEncoder->m_top - m_pAVIEncoder->m_bottom;

	if(crop_width>0 && crop_height>0)
	{
	    BYTE	 *pInData, *pOutData;   // Pointer to the actual image buffer
		BYTE	**ppInRows;			// the pointer to the input rows
		int		i, lineWidth;

		pSample->GetPointer(&pInData);
		pOutData=(BYTE *)malloc(sizeof(RGBTRIPLE)*m_pAVIEncoder->m_width*m_pAVIEncoder->m_height);

		// now let's do the cropping and scaling
		ppInRows=(BYTE **)malloc(sizeof(BYTE *)*crop_height);

		lineWidth=sizeof(RGBTRIPLE)*pvi->bmiHeader.biWidth;
		pInData+=(lineWidth*m_pAVIEncoder->m_bottom)+(sizeof(RGBTRIPLE)*m_pAVIEncoder->m_left);	// NOTE: image is top down !!!

		// now create an array with the input lines - cropping is done :-)
		if(m_pAVIEncoder->m_deinterlace!=0)
		{
			crop_height/=2;
			lineWidth*=2;
		}

		for(i=0; i<crop_height; i++)
		{
			ppInRows[i]=pInData;
			pInData+=lineWidth;
		}

		// now let's do the scaling
		rescale(ppInRows, crop_width, crop_height, m_pAVIEncoder->m_width, m_pAVIEncoder->m_height, pOutData);
/*		int	outline=sizeof(RGBTRIPLE)*m_width;
		for(i=0; i<crop_height; i++)
		{
			memcpy(pOutData, ppInRows[i], sizeof(RGBTRIPLE)*crop_width);
			pOutData+=outline;
		}
*/

		free(ppInRows);

		if(m_pAVIEncoder->m_aviinit==0)
		{
			LPBITMAPINFOHEADER	lpbi;
			DWORD				rate, scale;

			lpbi=(LPBITMAPINFOHEADER)calloc(sizeof(BITMAPINFOHEADER), 1);
			lpbi->biSize          = sizeof(BITMAPINFOHEADER);
			lpbi->biWidth         = m_pAVIEncoder->m_width;
			lpbi->biHeight        = m_pAVIEncoder->m_height;
			lpbi->biPlanes        = 1;
			lpbi->biBitCount      = 24;
			lpbi->biXPelsPerMeter = 0;
			lpbi->biYPelsPerMeter = 0;
			lpbi->biCompression   = BI_RGB;
    		lpbi->biSizeImage     = sizeof(RGBTRIPLE)*m_pAVIEncoder->m_width*m_pAVIEncoder->m_height;

			scale=10000;
			if(m_pAVIEncoder->m_frate<=0.0)
			{
				REFERENCE_TIME      fpms;

				fpms=10000000*(REFERENCE_TIME)scale;
				fpms/=pvi->AvgTimePerFrame;

				rate=(DWORD)fpms;

			}
			else
				rate=(DWORD)((double)(m_pAVIEncoder->m_frate*(double)scale));

			hr=initAVI(m_pAVIEncoder->m_filename, lpbi, rate, scale, m_pAVIEncoder->m_aviopts);
			free(lpbi);

			if(FAILED(hr))
			{
				m_pAVIEncoder->m_aviinit=2;

				free(pOutData);
				return CO_E_FAILEDTOCREATEFILE;
			}

			m_pAVIEncoder->m_aviinit=1;
		}

		writeAVI(pOutData, sizeof(RGBTRIPLE)*m_pAVIEncoder->m_width*m_pAVIEncoder->m_height);

		free(pOutData);
	}


    return NOERROR;
}


//
// EndOfStream
//
STDMETHODIMP CAVIEncoderInputPin::EndOfStream(void)
{
    CAutoLock lock(m_pReceiveLock);
    return CRenderedInputPin::EndOfStream();

} // EndOfStream


//
// NewSegment
//
// Called when we are seeked
//
STDMETHODIMP CAVIEncoderInputPin::NewSegment(REFERENCE_TIME tStart,
                                       REFERENCE_TIME tStop,
                                       double dRate)
{
    m_tLast = 0;
    return S_OK;

} // NewSegment


//
//  CAVIEncoder class
//
CAVIEncoder::CAVIEncoder(LPUNKNOWN pUnk, HRESULT *phr) :
    CUnknown(NAME("CAVIEncoder"), pUnk),
    m_pFilter(NULL),
    m_pPin(NULL),
    m_pPosition(NULL),
	m_top(0), m_bottom(0), m_left(0), m_right(0), m_width(416), m_height(224), m_deinterlace(0), m_frate(0.0), m_aviopts(1),
	m_startFrame(0), m_stopFrame(-1), m_aviinit(0)
{
	_tcscpy(m_filename, TEXT("outfile.avi"));

    m_pFilter = new CAVIEncoderFilter(this, GetOwner(), &m_Lock, phr);
    if (m_pFilter == NULL) {
        *phr = E_OUTOFMEMORY;
        return;
    }

    m_pPin = new CAVIEncoderInputPin(this,GetOwner(),
                               m_pFilter,
                               &m_Lock,
                               &m_ReceiveLock,
                               phr);
    if (m_pPin == NULL) {
        *phr = E_OUTOFMEMORY;
        return;
    }
}

// Destructor

CAVIEncoder::~CAVIEncoder()
{
    closeAVI();
	m_aviinit=0;
    delete m_pPin;
    delete m_pFilter;
    delete m_pPosition;
}


//
// CreateInstance
//
// Provide the way for COM to create a AVIEncoder filter
//
CUnknown * WINAPI CAVIEncoder::CreateInstance(LPUNKNOWN punk, HRESULT *phr)
{
    CAVIEncoder *pNewObject = new CAVIEncoder(punk, phr);
    if (pNewObject == NULL) {
        *phr = E_OUTOFMEMORY;
    }
    return pNewObject;

} // CreateInstance


//
// NonDelegatingQueryInterface
//
// Override this to say what interfaces we support where
//
STDMETHODIMP CAVIEncoder::NonDelegatingQueryInterface(REFIID riid, void ** ppv)
{
    CheckPointer(ppv,E_POINTER);
    CAutoLock lock(&m_Lock);

    // Do we have this interface

    if (riid == IID_IIPAVIEnc) {
        return GetInterface((IIPAVIEnc *) this, ppv);
    } else if (riid == IID_ISpecifyPropertyPages) {
        return GetInterface((ISpecifyPropertyPages *) this, ppv);
    } else if (riid == IID_IBaseFilter || riid == IID_IMediaFilter || riid == IID_IPersist) {
    	return m_pFilter->NonDelegatingQueryInterface(riid, ppv);
    } else if (riid == IID_IMediaPosition || riid == IID_IMediaSeeking) {
        if (m_pPosition == NULL) {

            HRESULT hr = S_OK;
            m_pPosition = new CPosPassThru(NAME("AVIEncoder Pass Through"),
                                           (IUnknown *) GetOwner(),
                                           (HRESULT *) &hr, m_pPin);
            if (m_pPosition == NULL) {
                return E_OUTOFMEMORY;
            }

            if (FAILED(hr)) {
                delete m_pPosition;
                m_pPosition = NULL;
                return hr;
            }
        }
        return m_pPosition->NonDelegatingQueryInterface(riid, ppv);
    } else {
	return CUnknown::NonDelegatingQueryInterface(riid, ppv);
    }

} // NonDelegatingQueryInterface


//
// GetPages
//
// Returns the clsid's of the property pages we support
//
STDMETHODIMP CAVIEncoder::GetPages(CAUUID *pPages)
{
    pPages->cElems = 1;
    pPages->pElems = (GUID *) CoTaskMemAlloc(sizeof(GUID));
    if (pPages->pElems == NULL) {
        return E_OUTOFMEMORY;
    }
    *(pPages->pElems) = CLSID_AVIEncoderPropertyPage;
    return NOERROR;

} // GetPages


//
// get_IPScaleCrop
//
// get current parameters
//
STDMETHODIMP CAVIEncoder::get_IPAVIEnc(int *top, int *bottom, int *left, int *right, int *width, int *height, int *deint, double *frate, LPTSTR ptstr, int *showopts)             
{
    CAutoLock cAutolock(&m_Lock);
    CheckPointer(top,E_POINTER);
    CheckPointer(bottom,E_POINTER);
    CheckPointer(left,E_POINTER);
    CheckPointer(right,E_POINTER);
    CheckPointer(width,E_POINTER);
    CheckPointer(height,E_POINTER);
    CheckPointer(deint,E_POINTER);

	*top=m_top;
	*bottom=m_bottom;
	*left=m_left;
	*right=m_right;
	*width=m_width;
	*height=m_height;
	*frate=m_frate;
	*deint=m_deinterlace;
	*showopts=m_aviopts;

	_tcscpy(ptstr, m_filename);

    return NOERROR;

} // get_IPEffect


//
// put_IPScaleCrop
//
// set parameters
//
STDMETHODIMP CAVIEncoder::put_IPAVIEnc(int top, int bottom, int left, int right, int width, int height, int deint, double frate, LPTSTR ptstr, int showopts)               
{
    CAutoLock cAutolock(&m_Lock);

	m_top=top;
	m_bottom=bottom;
	m_left=left;
	m_right=right;
	m_width=width;
	m_height=height;
	m_frate=frate;
	m_deinterlace=deint;
	m_aviopts=showopts;

	_tcscpy(m_filename, ptstr);

    return NOERROR;

} // put_IPEffect

STDMETHODIMP CAVIEncoder::get_FrameCount(long *frames)
{
    CAutoLock cAutolock(&m_Lock);

	*frames=frameNr;

    return NOERROR;
}

STDMETHODIMP CAVIEncoder::close_AVI()
{
    CAutoLock cAutolock(&m_Lock);
	CAutoLock cAutoReceivelock(&m_ReceiveLock);

	closeAVI();

    return NOERROR;
}


#define SCALE 4096
#define HALFSCALE 2048

typedef struct {
  unsigned long red;
  unsigned long green;
  unsigned long blue;
} colorsum;

void multiply(unsigned int x,unsigned int *p)
{
  int i  =  256;
  unsigned int k = 0;

  while (--i >= 0) {
    *p++ = k;
    k += x;
  }
}

void rescale(unsigned char **inpic, int oldw, int oldh,int neww, int newh, unsigned char *outpic)
{
  unsigned char  *xelrow;
  unsigned char  *tempxelrow;
  unsigned char  *newxelrow;
  unsigned char  *rb2;
  unsigned char *xP;
  unsigned char *nxP;
  int             rowsread;
  int    row, col, needtoreadrow;
  long            sxscale, syscale;
  long   fracrowtofill, fracrowleft;
  colorsum       *cs;
  unsigned int mult[256];
  unsigned int multcol[256];

  sxscale = neww * SCALE;
  sxscale /= oldw;
  multiply(sxscale,multcol);
  syscale = newh * SCALE;
  syscale /= oldh;

  cs = (colorsum *) malloc(oldw * sizeof(colorsum));
  rowsread = 0;
  fracrowleft = syscale;
  needtoreadrow = 1;
  for (col = 0; col < oldw; ++col)
    cs[col].red = cs[col].green = cs[col].blue = HALFSCALE;
  fracrowtofill = SCALE;

  rb2 = (unsigned char *) malloc(3*oldw);

  xelrow = inpic[rowsread]; 
  for (row = 0; row < newh; ++row) {
    newxelrow = &outpic[row*neww*3];
    if (neww == oldw) {	/* No X scaling later */
      tempxelrow = newxelrow;
    } else {
      tempxelrow = rb2;
    }
    /* First scale Y from xelrow into tempxelrow. */
    if (newh != oldh) {	/* shortcut Y scaling if possible */
      while (fracrowleft < fracrowtofill) {
	if (needtoreadrow)
	  if (rowsread < oldh) {
	    xelrow = inpic[rowsread++];
	  }
	multiply(fracrowleft,mult);
	for (col = 0, xP = xelrow; col < oldw; ++col, xP+=3) {
	  cs[col].red += mult[xP[0]];
	  cs[col].green += mult[xP[1]];
	  cs[col].blue += mult[xP[2]];
	}
	fracrowtofill -= fracrowleft;
	fracrowleft = syscale;
	needtoreadrow = 1;
      }
      /* Now fracrowleft is >= fracrowtofill, so we can produce a row. */
      if (needtoreadrow)
	if (rowsread < oldh) {
	  xelrow = inpic[rowsread++];
	  needtoreadrow = 0;
	}
      multiply(fracrowtofill,mult);
      for (col = 0, xP = xelrow, nxP = tempxelrow;
	   col < oldw; ++col, xP += 3, nxP += 3)  {
	register long   r, g, b;

	r = cs[col].red + mult[xP[0]];
	g = cs[col].green + mult[xP[1]];
	b = cs[col].blue + mult[xP[2]];
	r /= SCALE;
	if (r > 255)
	  r = 255;
	g /= SCALE;
	if (g > 255)
	  g = 255;
	b /= SCALE;
	if (b > 255)
	  b = 255;
	nxP[0] = (unsigned char)r;
	nxP[1] = (unsigned char)g;
	nxP[2] = (unsigned char)b;
	cs[col].red = cs[col].green = cs[col].blue = HALFSCALE;
      }
      fracrowleft -= fracrowtofill;
      if (fracrowleft == 0) {
	fracrowleft = syscale;
	needtoreadrow = 1;
      }
      fracrowtofill = SCALE;
    }

    /* Now scale X from tempxelrow into newxelrow and write it out. */
    if (neww != oldw) {	/* shortcut X scaling if possible */
      register long   r, g, b;
      register long   fraccoltofill, fraccolleft;
      register int    needcol;

      nxP = newxelrow;
      fraccoltofill = SCALE;
      r = g = b = HALFSCALE;
      needcol = 0;
      for (col = 0, xP = tempxelrow; col < oldw; ++col, xP += 3) {
	fraccolleft = sxscale;
	while (fraccolleft >= fraccoltofill) {
	  if (needcol) {
	    nxP += 3;
	    r = g = b = HALFSCALE;
	  }
	  r += fraccoltofill * xP[0];
	  g += fraccoltofill * xP[1];
	  b += fraccoltofill * xP[2];
	  r /= SCALE;
	  if (r > 255)
	    r = 255;
	  g /= SCALE;
	  if (g > 255)
	    g = 255;
	  b /= SCALE;
	  if (b > 255)
	    b = 255;
	  nxP[0] = (unsigned char)r;
	  nxP[1] = (unsigned char)g;
	  nxP[2] = (unsigned char)b;
	  fraccolleft -= fraccoltofill;
	  fraccoltofill = SCALE;
	  needcol = 1;
	}
	if (fraccolleft > 0) {
	  if (needcol) {
	    nxP += 3;
	    r = g = b = HALFSCALE;
	    needcol = 0;
	  }
	  if (fraccolleft == sxscale) {
	    r += multcol[xP[0]];
	    g += multcol[xP[1]];
	    b += multcol[xP[2]];
          } else {
	    r += fraccolleft * xP[0];
	    g += fraccolleft * xP[1];
	    b += fraccolleft * xP[2];
	  }
	  fraccoltofill -= fraccolleft;
	}
      }
      if (fraccoltofill > 0) {
	xP -= 3;
	r += fraccoltofill * xP[0];
	g += fraccoltofill * xP[1];
	b += fraccoltofill * xP[2];
      }
      if (!needcol) {
	r /= SCALE;
	if (r > 255)
	  r = 255;
	g /= SCALE;
	if (g > 255)
	  g = 255;
	b /= SCALE;
	if (b > 255)
	  b = 255;
	nxP[0] = (unsigned char)r;
	nxP[1] = (unsigned char)g;
	nxP[2] = (unsigned char)b;
      }
    }
  }
  free(cs);
  free(rb2);
}

