#include <windows.h>
#include <initguid.h>
#include <streams.h>
#if (1100 > _MSC_VER)
#include <olectlid.h>
#else
#include <olectl.h>
#endif
#include <stdlib.h>
#include <stdio.h>

#include "ave_uids.h"
#include "inf_ave.h"
#include "mfs_uids.h"
#include "inf_mfs.h"
#include "bndpuids.h"

static int cleft=0;
static int cright=0;
static int ctop=0;
static int cbottom=0;
static int swidth=480;
static int sheight=208;
static int deint=0;
static double frate=25.0;
static long frames=-1;
static int aviopts=1;
static LONGLONG	blocksize=2048;
static LONGLONG	skip=0;
static LONGLONG	start=0;
static LONGLONG	stop=0;
static int		range=0;

static MFS_FileList *infiles=NULL;
static char	*outfile="output.avi";
static char *audiobase=NULL;

static IIPAVIEnc		*pAVIEnc=NULL;			// a pointer to my filters interface
static IGraphBuilder	*pGraph=NULL;			// a pointer to my graphs interface

static BOOL WINAPI myHandler(DWORD dwCtrlType);
static void printUsage(char *cmd);
static HRESULT renderVideo();
static HRESULT playVideo();
static void closeVideo();
static void waitVideoToFinish();
static LONGLONG atoll(char *s);

int main(int argc, char	**argv)
{
	MFS_FileList	*fl, *pfl;
	HRESULT	hr;
	int		argn;

	argn=1;

	while(argn<argc && argv[argn][0]=='-')
	{
		switch(argv[argn][1])
		{
			case 'a':
				if(strcmp(argv[argn], "-audio")==0)
				{
					audiobase=argv[++argn];
				}
				else if(argv[argn][2]==0)
					aviopts=0;
				else
					goto usage;

				break;

			case 'b':
				if(strcmp(argv[argn], "-bs")==0)
				{
					blocksize=atoll(argv[++argn]);
				}
				else
					goto usage;
				break;

			case 'd':
				if(argv[argn][2]!=0)
					goto usage;

				deint=1;
				break;

			case 'o':
				if(argv[argn][2]!=0)
					goto usage;

				outfile=argv[++argn];
				break;

			case 'n':
				if(argv[argn][2]!=0)
					goto usage;

				frames=atoi(argv[++argn]);
				break;

			case 'r':
				if(strcmp(argv[argn], "-range")==0)
				{
					range=1;
					start=atoll(argv[++argn]);
					argn++;

					if(strcmp(argv[argn], "-1")==0)
						stop=-1;
					else
						stop=atoll(argv[argn]);
				}
				else if(strcmp(argv[argn], "-ranget")==0)
				{
					range=2;
					start=atoll(argv[++argn])*10000;	// to get it in 100ns resolution
					argn++;

					if(strcmp(argv[argn], "-1")==0)
						stop=-1;
					else
						stop=atoll(argv[argn])*10000;	// to get it in 100ns resolution
				}
				else if(argv[argn][2]==0)
					frate=atof(argv[++argn]);
				else
					goto usage;

				break;

			case 's':
				if(strcmp(argv[argn], "-skip")==0)
				{
					skip=atoll(argv[++argn]);
				}
				else
					goto usage;
				break;

			default:
				goto usage;
		}

		argn++;
	}

	if(argc-argn < 7)
	{
usage:
		printUsage(argv[0]);
		exit(0);
	}

	cleft=atoi(argv[argn++]);
	cright=atoi(argv[argn++]);
	ctop=atoi(argv[argn++]);
	cbottom=atoi(argv[argn++]);
	swidth=atoi(argv[argn++]);
	sheight=atoi(argv[argn++]);

	skip*=blocksize;

	// let's create the infiles list
	while(argn<argc)
	{
		int	len;

		len=strlen(argv[argn]);
		fl=(MFS_FileList *)CoTaskMemAlloc(sizeof(MFS_FileList)+sizeof(WCHAR)*len);
		if(fl)
		{
			fl->Next=NULL;

			// now let's render the input file
			MultiByteToWideChar(CP_ACP, 0, argv[argn], -1, fl->Filename, (len+1));
		
			if(infiles==NULL)
				infiles=fl;
			else
				pfl->Next=fl;

			pfl=fl;
		}

		argn++;
	}

	if(infiles==NULL)
		goto usage;

	SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);

	// first disable default Control-C handling and install my handler
	SetConsoleCtrlHandler(myHandler, TRUE);	

	CoInitialize(NULL);
	hr=CoCreateInstance(CLSID_FilterGraph,          // get this documents graph object
						NULL,
						CLSCTX_INPROC_SERVER,
						IID_IGraphBuilder,
						(void **)&pGraph);
	if(FAILED(hr))
	{
		printf("Unable to initialize ActiveMovie !\nPlease make sure Microsoft ActiveMovie is installed.\n");
		goto exit;
	}

	hr=renderVideo();
	if(FAILED(hr))
	{
		printf("failed to render video - check infiles\n");
		goto exit;
	}

	hr=playVideo();
	if(FAILED(hr))
	{
		printf("failed to play video\n");
		goto exit;
	}

	// in case avi encoder changed our priority again
	SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);

	waitVideoToFinish();

exit:
	if(pGraph)
	{
		closeVideo();

		pGraph->Release();
	}

	return 0;
}

static BOOL WINAPI myHandler(DWORD dwCtrlType)
{
	if(pGraph)
	{
		closeVideo();

		pGraph->Release();
	}

	exit(0);
}
 
static void printUsage(char *cmd)
{
	printf("AVIEncoder v0.20\n");
	printf("Usage: %s [-a] [-n framestoencode] [-range[t] start stop] [-skip count] [-bs blocksize] [-r framerate] [-d] [-audio filebase] [-o outfile] left right top bottom width height infile1 [infile2 ...]\n", cmd);
    printf("  -a: don't show avi compression options (use previous parameters)\n");
	printf("  -n: stop after this number of frames have been encoded\n");
	printf("  -range: encode this frame range (a end frame of -1 for end of movie)\n"); 
	printf("  -ranget: encode this time range (in millisecs, -1 for end of movie)\n"); 
	printf("  -skip: skip the given number of blocks at the beginning of the file\n");
	printf("  -bs: block size in bytes to use for skipping (default 2048 bytes)\n");
	printf("  -r: set framerate for output avi file (NTSC: 23.976 or 24.0)\n");
	printf("  -d: deinterlace\n");
	printf("  -audio: write audio streams to files (.ac3, .mpa or .unk is added)\n");
	printf("  resolutions: 4:3 = 384x288, 16:9 = 416x224, 2.35:1 = 480x208\n"); 
}

static HRESULT renderVideo()
{
	IEnumFilters	*pEF;
	IBaseFilter		*pBF, *pBF_AVI;
	ULONG			fetched;
	IEnumPins		*pEP;
	IPin			*pIn, *pOut;
	HRESULT			hr;

	// let's create an instance of the AVIEncoder
	hr=CoCreateInstance(CLSID_AVIEncoder, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void **)&pBF);
	if(FAILED(hr))
		return hr;

	hr=pGraph->AddFilter(pBF, NULL);
    if(FAILED(hr))
	{
		pBF->Release();
		return hr;
	}

	// now let's find the IIPAVIEnc interface
	hr=pBF->QueryInterface(IID_IIPAVIEnc, (void **)&pAVIEnc);
    if(FAILED(hr))
	{
		pBF->Release();
		return hr;
	}
	pBF_AVI=pBF;
	pBF=NULL;

	// now let's set the parameters for the encoder
	pAVIEnc->put_IPAVIEnc(ctop, cbottom, cleft, cright, swidth, sheight, deint, frate, outfile, aviopts);

	hr=pGraph->RenderFile(infiles->Filename, NULL);
	if(FAILED(hr))
	{
		pBF_AVI->Release();
		return hr;
	}

	if(infiles->Next!=NULL || skip!=0)		// more than one infile, so let's use our filter
	{
		IIPMFSrc			*pMFS;
		IPin				*pPN[2];
		PIN_DIRECTION		pd;
		AM_MEDIA_TYPE		mType;
		DWORD				i;
		BOOL				found=FALSE;

		// first let's find the old source filter
		hr=pGraph->EnumFilters(&pEF);
		if(FAILED(hr))
		{
			pBF_AVI->Release();
			return hr;
		}

		for(;;)
		{
			hr=pEF->Next(1, &pBF, &fetched);

			if(FAILED(hr) || fetched==0)
				break;

			hr=pBF->EnumPins(&pEP);
			if(SUCCEEDED(hr))
			{
				pEP->Reset();

				hr=pEP->Next(2, pPN, &fetched);
				if(fetched==1)		// all pins checked !
				{
					hr=pPN[0]->QueryDirection(&pd);
					if(SUCCEEDED(hr) && pd==PINDIR_OUTPUT)	// this is the source filter !!!!!
					{
						pPN[0]->ConnectionMediaType(&mType);					
						pGraph->RemoveFilter(pBF);
						found=TRUE;
					}
				}

				for(i=0; i<fetched; i++)
					pPN[i]->Release();

				pEP->Release();
			}

			pBF->Release();

			if(found)
				break;
		}

		pEF->Release();

		if(found==FALSE)	// we didn't file where the source filter was connected to :-(
		{
			pBF_AVI->Release();
			return S_FALSE;
		}

		// now we insert our multiple file source filter

		hr=CoCreateInstance(CLSID_MultFileSrc, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void **)&pBF);
		if(FAILED(hr))
		{
			pBF_AVI->Release();
			return hr;
		}

		hr=pGraph->AddFilter(pBF, NULL);
		if(FAILED(hr))
		{
			pBF->Release();
			pBF_AVI->Release();
			return hr;
		}

		// now let's find the IIPMFSrc interface
		hr=pBF->QueryInterface(IID_IIPMFSrc, (void **)&pMFS);
		if(FAILED(hr))
		{
			pBF->Release();
			pBF_AVI->Release();
			return hr;
		}

		// now put media type and file list
		hr=pMFS->put_MediaType(&mType);
		if(FAILED(hr))
		{
			pMFS->Release();
			pBF->Release();
			pBF_AVI->Release();
			return hr;
		}

		hr=pMFS->put_FileList(infiles, skip);
		if(FAILED(hr))
		{
			pMFS->Release();
			pBF->Release();
			pBF_AVI->Release();
			return hr;
		}

		pMFS->Release();

		// now connect source filter with the graph
		hr=pBF->FindPin(L"Output", &pOut);
		if(FAILED(hr))
		{
			pBF->Release();
			pBF_AVI->Release();
			return hr;
		}

		// render our output pin
		hr=pGraph->Render(pOut);

		pBF->Release();
		pOut->Release();

		if(FAILED(hr))
		{
			pBF_AVI->Release();
			return hr;
		}
	}

	pBF=pBF_AVI;
	pBF_AVI=NULL;

	// now check if our filter is really a part of the graph
	hr=pBF->EnumPins(&pEP);
	if(FAILED(hr))
	{
		pBF->Release();
		return hr;
	}

	pBF->Release();

	pEP->Reset();
	pIn=NULL;
	hr=pEP->Next(1, &pIn, &fetched);
	if(FAILED(hr))
	{
		pEP->Release();
		return hr;
	}
	pEP->Release();

	if(fetched!=1)
	{
		return S_FALSE;
	}

	hr=pIn->ConnectedTo(&pOut);
	if(FAILED(hr))
	{
		DWORD		i;
		// we're not in graph, so let's find the video renderer and replace us with him
		hr=pGraph->EnumFilters(&pEF);
		if(FAILED(hr))
		{
			pIn->Release();
			return hr;
		}

		pOut=NULL;
		pEF->Reset();
		while(pOut==NULL)
		{
			hr=pEF->Next(1, &pBF, &fetched);

			if(FAILED(hr) || fetched==0)
				break;

			hr=pBF->QueryInterface(IID_IIPAVIEnc, (void **)&pAVIEnc);	// is this our own filter ?
			if(SUCCEEDED(hr))	
			{
				pAVIEnc->Release();
				pBF->Release();
				continue;
			}

			hr=pBF->EnumPins(&pEP);
			if(SUCCEEDED(hr))
			{
				IPin			*pPN[2];

				pEP->Reset();

				hr=pEP->Next(2, pPN, &fetched);
				if(fetched==1)		// all pins checked !
				{
					PIN_DIRECTION	pd;

					hr=pPN[0]->QueryDirection(&pd);
					if(SUCCEEDED(hr) && pd==PINDIR_INPUT)
					{
						AM_MEDIA_TYPE	mType;

						hr=pPN[0]->ConnectionMediaType(&mType);					
						if(SUCCEEDED(hr) && IsEqualGUID(mType.majortype, MEDIATYPE_Video))
						{
							hr=pPN[0]->ConnectedTo(&pOut);
							if(FAILED(hr))
								pOut=NULL;
							else
								pGraph->RemoveFilter(pBF);
						}	
					}
				}

				for(i=0; i<fetched; i++)
					pPN[i]->Release();

				pEP->Release();
			}

			pBF->Release();
		}

		pEF->Release();

		if(pOut==NULL)
		{
			pIn->Release();
			return S_FALSE;
		}

		hr=pGraph->Connect(pOut, pIn);
		if(FAILED(hr))
		{
			pIn->Release();
			pOut->Release();

			return hr;
		}
	}

	pIn->Release();
	pOut->Release();

	// now let's remove all audio filters from graph
	hr=pGraph->EnumFilters(&pEF);
	if(FAILED(hr))
		return hr;

	pEF->Reset();
	for(;;)
	{
		hr=pEF->Next(1, &pBF, &fetched);

		if(FAILED(hr) || fetched==0)
			break;

		hr=pBF->EnumPins(&pEP);
		if(SUCCEEDED(hr))
		{
			AM_MEDIA_TYPE	mType;
			IPin			*pPN;

			pEP->Reset();

			for(;;)
			{
				hr=pEP->Next(1, &pPN, &fetched);
				if(FAILED(hr) || fetched==0)		// all pins checked !
				{
					pGraph->RemoveFilter(pBF);
					pEF->Reset();

					break;
				}

				hr=pPN->ConnectionMediaType(&mType);					
				pPN->Release();

				if(SUCCEEDED(hr))
				{
					if(!IsEqualGUID(mType.majortype, MEDIATYPE_Audio))	// is this a not audio pin - so we don't remove this filter?
						break;	
				}
			}

			pEP->Release();
		}

		pBF->Release();
	}

	pEF->Release();

	/* does the user want us to dump the audio streams? */
	if(audiobase!=NULL)
	{
		int		audiocount=0;

		hr=pGraph->EnumFilters(&pEF);
		if(FAILED(hr))
			return hr;

		pEF->Reset();
		for(;;)
		{
			hr=pEF->Next(1, &pBF, &fetched);

			if(FAILED(hr) || fetched==0)
				break;

			hr=pBF->EnumPins(&pEP);
			if(SUCCEEDED(hr))
			{
				IPin			*pPN, *pPN2;
				PIN_DIRECTION	pinDir;

				pEP->Reset();

				for(;;)
				{
					hr=pEP->Next(1, &pPN, &fetched);
					if(FAILED(hr) || fetched==0)		// all pins checked !
						break;

					pPN->QueryDirection(&pinDir);
					pPN->ConnectedTo(&pPN2);
					if(pPN2!=NULL)
						pPN2->Release();

					if(pPN2==NULL && pinDir==PINDIR_OUTPUT)
					{
						IEnumMediaTypes *pEMT;

						hr=pPN->EnumMediaTypes(&pEMT);
						if(SUCCEEDED(hr))
						{
							AM_MEDIA_TYPE	*pmt;
							pEMT->Reset();

							for(;;)
							{
								hr=pEMT->Next(1, &pmt, &fetched);
								if(FAILED(hr) || fetched==0)		// all pins checked !
									break;

								if(IsEqualGUID(pmt->majortype, MEDIATYPE_Audio))	// is this a not connected output audio pin?
								{
									IBaseFilter		*pBF2;
									char			*ext=".unk";
									char			outfile[MAX_PATH];

									if(IsEqualGUID(pmt->subtype, MEDIASUBTYPE_DOLBY_AC3))
										ext=".ac3";
									else if(IsEqualGUID(pmt->subtype, MEDIASUBTYPE_MPEG1Audio))
										ext=".mpa";

									sprintf(outfile, "%s%d%s", audiobase, audiocount, ext);

									// create a bindump filter and connect it to this pin
									hr=CoCreateInstance(CLSID_BinDump, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void **)&pBF2);
									if(SUCCEEDED(hr))
									{
										hr=pGraph->AddFilter(pBF2, NULL);
										if(SUCCEEDED(hr))
										{
											IFileSinkFilter	*pFS;

											// now let's find the IFileSinkFilter interface
											hr=pBF2->QueryInterface(IID_IFileSinkFilter, (void **)&pFS);
											if(SUCCEEDED(hr))
											{
												WCHAR	wname[MAX_PATH];

												MultiByteToWideChar(CP_ACP, 0, outfile, -1, wname, MAX_PATH);

												hr=pFS->SetFileName(wname, pmt);
												if(SUCCEEDED(hr))
												{
													// now connect the input pin
													hr=pBF2->FindPin(L"Input", &pPN2);
													if(SUCCEEDED(hr))
													{
														hr=pPN->Connect(pPN2, pmt);
														if(SUCCEEDED(hr))
															printf("AUDIOSTREAM %d: %s\n", audiocount, outfile);

														pPN2->Release();
													}
												}

												pFS->Release();
											}
										}
										
										pBF2->Release();
									}

									audiocount++;
									DeleteMediaType(pmt);
									break;
								}

								DeleteMediaType(pmt);
							}

							pEMT->Release();
						}
					}

					pPN->Release();
				}

				pEP->Release();
			}

			pBF->Release();
		}

		pEF->Release();

	}

	return S_OK;
}

static HRESULT playVideo()
{
	HRESULT			hr;
	IMediaControl	*pMC;

	if(range!=0)
	{
		IMediaSeeking	*pMS;

		hr=pGraph->QueryInterface(IID_IMediaSeeking, (void **)&pMS);
		if(SUCCEEDED(hr))
		{
			if(range==1)
				hr=pMS->SetTimeFormat(&TIME_FORMAT_FRAME);
			else if(range==2)
				hr=pMS->SetTimeFormat(&TIME_FORMAT_MEDIA_TIME);

			if(SUCCEEDED(hr))
			{
				if(stop==-1)
					hr=pMS->SetPositions(&start, AM_SEEKING_AbsolutePositioning, &stop, AM_SEEKING_NoPositioning);
				else
					hr=pMS->SetPositions(&start, AM_SEEKING_AbsolutePositioning, &stop, AM_SEEKING_AbsolutePositioning);

				if(FAILED(hr))
					fprintf(stderr, "FAILED to set encoding range - failed to set range: %d!\n", hr);
			}
			else
				fprintf(stderr, "FAILED to set encoding range - unable to set seeking timeformat: %d!\n", hr);

			pMS->Release();
		}
		else
			fprintf(stderr, "FAILED to set encoding range - unable to get IMediaSeeking interface: %d!\n", hr);
	}

	// Obtain the interface to our filter graph
	hr=pGraph->QueryInterface(IID_IMediaControl, (void **)&pMC);
	if(SUCCEEDED(hr))
	{
		hr=pMC->Run();

		pMC->Release();
	}

	return hr;
}

static void closeVideo()
{
	HRESULT					hr;
	IMediaControl			*pMC;
	IBaseFilter				*pFilter;
	IEnumFilters			*pEnumFilter;
	ULONG					pcFetched;

	if(pAVIEnc!=NULL)		// in case the graph is in trouble, we close avi first !!!
	{
		pAVIEnc->close_AVI();
		pAVIEnc->Release();
	}

	// Obtain the interface to our filter graph
	hr=pGraph->QueryInterface(IID_IMediaControl, (void **)&pMC);
	if(SUCCEEDED(hr))
	{
		hr=pMC->Stop();		// only to be sure

		pMC->Release();
	}

	hr=pGraph->EnumFilters(&pEnumFilter);
	if(SUCCEEDED(hr))
	{
		for(;;)
		{
			pEnumFilter->Reset();
		
			hr=pEnumFilter->Next(1, &pFilter, &pcFetched);

			if(FAILED(hr) || pcFetched==0)
				break;

			pGraph->RemoveFilter(pFilter);

			pFilter->Release();
		}

		pEnumFilter->Release();
	}	
}

static void waitVideoToFinish()
{
	HRESULT		hr;
	IMediaEvent	*pME;

	// Obtain the interface to our filter graph
	hr=pGraph->QueryInterface(IID_IMediaEvent, (void **)&pME);
	if(SUCCEEDED(hr))
	{
		long	evt, lfr=-1, fr, lParam1, lParam2;
		int		poll;

		poll=100;		// default is to poll every 100ms the new frame number
		if(frames>0)
			poll=10;	// if we stop the stream let's poll a bit more :-)

		for(;;)
		{	
			hr=pAVIEnc->get_FrameCount(&fr);
			if(lfr!=fr)
			{
				printf("\rframes encoded: %d", fr);
				fflush(stdout);
			}
			lfr=fr;

			if(SUCCEEDED(hr) && fr>=frames && frames>0)
			{
				break;
			}

			hr=pME->GetEvent(&evt, &lParam1, &lParam2, poll);	// we poll every 50ms
			if(SUCCEEDED(hr))
			{
				pME->FreeEventParams(evt, lParam1, lParam2);

				if(evt==EC_COMPLETE || evt==EC_ERRORABORT || evt==EC_USERABORT)
					break;
			}
		}

		pME->Release();
	}
}

static LONGLONG atoll(char *s)
{
	LONGLONG		ll;
	unsigned int	i;

	ll=0;
	for(i=0; i<strlen(s); i++)
	{
		ll*=10;
		ll+=(s[i]-'0');
	}

	return ll;
}

void WINAPI DeleteMediaType(AM_MEDIA_TYPE *pmt)
{
    // allow NULL pointers for coding simplicity

    if (pmt == NULL) {
        return;
    }

    FreeMediaType(*pmt);
    CoTaskMemFree((PVOID)pmt);
}


void WINAPI FreeMediaType(AM_MEDIA_TYPE& mt)
{
    if (mt.cbFormat != 0) {
        CoTaskMemFree((PVOID)mt.pbFormat);

        // Strictly unnecessary but tidier
        mt.cbFormat = 0;
        mt.pbFormat = NULL;
    }
    if (mt.pUnk != NULL) {
        mt.pUnk->Release();
        mt.pUnk = NULL;
    }
}

