#include <windows.h>
#include <windowsx.h>
#include <streams.h>
#include <commctrl.h>
#include <olectl.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <tchar.h>
#include "resource.h"
#include "mfs_uids.h"
#include "inf_mfs.h"
#include "multfilesrc.h"
#include "mfs_prop_media.h"


// this are the media type definitions taken from uuids.h and ksuuids.h

typedef struct
{
    const CLSID __RPC_FAR	*type;
	LPSTR					name;
} MTYPE_ENTRY;

static MTYPE_ENTRY	major_types[] =
{
	{ &MEDIATYPE_NULL				, "NULL"				},
	{ &MEDIATYPE_Video				, "Video"				},
	{ &MEDIATYPE_Audio				, "Audio"				},
	{ &MEDIATYPE_Text				, "Text"				},
	{ &MEDIATYPE_Midi				, "Midi"				},
	{ &MEDIATYPE_Stream				, "Stream"				},
	{ &MEDIATYPE_Interleaved		, "Interleaved"			},
	{ &MEDIATYPE_File				, "File"				},
	{ &MEDIATYPE_ScriptCommand		, "ScriptCommand"		},
	{ &MEDIATYPE_AUXLine21Data		, "AUXLine21Data"		},
	{ &MEDIATYPE_Timecode			, "Timecode"			},
	{ &MEDIATYPE_LMRT				, "LMRT"				},
	{ &MEDIATYPE_URL_STREAM			, "URL_STREAM"			},
	{ &MEDIATYPE_MPEG1SystemStream	, "MPEG1SystemStream"	},
	{ &MEDIATYPE_MPEG2_PES			, "MPEG2_PES"			},
	{ &MEDIATYPE_MPEG2_PACK			, "MPEG2_PACK"			},
	{ &MEDIATYPE_DVD_ENCRYPTED_PACK	, "DVD_ENCRYPTED_PACK"	},
	{ &MEDIATYPE_DVD_NAVIGATION		, "DVD_NAVIGATION"		},
};

static MTYPE_ENTRY	minor_types[] =
{
	{ &MEDIASUBTYPE_NULL				, "NULL"				},
	{ &MEDIASUBTYPE_None				, "None"				},
	{ &MEDIASUBTYPE_YVU9				, "YVU9"				},
	{ &MEDIASUBTYPE_Y411				, "Y411"				},
	{ &MEDIASUBTYPE_Y41P				, "Y41P"				},
	{ &MEDIASUBTYPE_YUY2				, "YUY2"				},
	{ &MEDIASUBTYPE_YVYU				, "YVYU"				},
	{ &MEDIASUBTYPE_UYVY				, "UYVY"				},
	{ &MEDIASUBTYPE_Y211				, "Y211"				},
	{ &MEDIASUBTYPE_YV12				, "YV12"				},
	{ &MEDIASUBTYPE_CLJR				, "CLJR"				},
	{ &MEDIASUBTYPE_IF09				, "IF09"				},
	{ &MEDIASUBTYPE_CPLA				, "CPLA"				},
	{ &MEDIASUBTYPE_MJPG				, "MJPG"				},
	{ &MEDIASUBTYPE_TVMJ				, "TVMJ"				},
	{ &MEDIASUBTYPE_WAKE				, "WAKE"				},
	{ &MEDIASUBTYPE_CFCC				, "CFCC"				},
	{ &MEDIASUBTYPE_IJPG				, "IJPG"				},
	{ &MEDIASUBTYPE_Plum				, "Plum"				},
	{ &MEDIASUBTYPE_DVCS				, "DVCS"				},
	{ &MEDIASUBTYPE_DVSD				, "DVSD"				},
	{ &MEDIASUBTYPE_MDVF				, "MDVF"				},
	{ &MEDIASUBTYPE_RGB1				, "RGB1"				},
	{ &MEDIASUBTYPE_RGB4				, "RGB4"				},
	{ &MEDIASUBTYPE_RGB8				, "RGB8"				},
	{ &MEDIASUBTYPE_RGB565				, "RGB565"				},
	{ &MEDIASUBTYPE_RGB555				, "RGB555"				},
	{ &MEDIASUBTYPE_RGB24				, "RGB24"				},
	{ &MEDIASUBTYPE_RGB32				, "RGB32"				},
	{ &MEDIASUBTYPE_Overlay				, "Overlay"				},
	{ &MEDIASUBTYPE_MPEG1Packet			, "MPEG1Packet"			},
	{ &MEDIASUBTYPE_MPEG1Payload		, "MPEG1Payload"		},
	{ &MEDIASUBTYPE_MPEG1AudioPayload	, "MPEG1AudioPayload"	},
	{ &MEDIASUBTYPE_MPEG1System			, "MPEG1System"			},
	{ &MEDIASUBTYPE_MPEG1VideoCD		, "MPEG1VideoCD"		},
	{ &MEDIASUBTYPE_MPEG1Video			, "MPEG1Video"			},
	{ &MEDIASUBTYPE_MPEG1Audio			, "MPEG1Audio"			},
	{ &MEDIASUBTYPE_Avi					, "Avi"					},
	{ &MEDIASUBTYPE_Asf					, "Asf"					},
	{ &MEDIASUBTYPE_QTMovie				, "QTMovie"				},
	{ &MEDIASUBTYPE_QTRpza				, "QTRpza"				},
	{ &MEDIASUBTYPE_QTSmc				, "QTSmc"				},
	{ &MEDIASUBTYPE_QTRle				, "QTRle"				},
	{ &MEDIASUBTYPE_QTJpeg				, "QTJpeg"				},
	{ &MEDIASUBTYPE_PCMAudio_Obsolete	, "PCMAudio_Obsolete"	},
	{ &MEDIASUBTYPE_PCM					, "PCM"					},
	{ &MEDIASUBTYPE_WAVE				, "WAVE"				},
	{ &MEDIASUBTYPE_AU					, "AU"					},
	{ &MEDIASUBTYPE_AIFF				, "AIFF"				},
	{ &MEDIASUBTYPE_dvsd				, "dvsd"				},
	{ &MEDIASUBTYPE_dvhd				, "dvhd"				},
	{ &MEDIASUBTYPE_dvsl				, "dvsl"				},
	{ &MEDIASUBTYPE_Line21_BytePair		, "Line21_BytePair"		},
	{ &MEDIASUBTYPE_Line21_GOPPacket	, "Line21_GOPPacket"	},
	{ &MEDIASUBTYPE_Line21_VBIRawData	, "Line21_VBIRawData"	},
	{ &MEDIASUBTYPE_DssVideo			, "DssVideo"			},
	{ &MEDIASUBTYPE_DssAudio			, "DssAudio"			},
	{ &MEDIASUBTYPE_VPVideo				, "VPVideo"				},
	{ &MEDIASUBTYPE_VPVBI				, "VPVBI"				},
	{ &MEDIASUBTYPE_MPEG2_PROGRAM		, "MPEG2_PROGRAM"		},
	{ &MEDIASUBTYPE_MPEG2_TRANSPORT		, "MPEG2_TRANSPORT"		},
	{ &MEDIASUBTYPE_MPEG2_VIDEO			, "MPEG2_VIDEO"			},
	{ &MEDIASUBTYPE_MPEG2_AUDIO			, "MPEG2_AUDIO"			},
	{ &MEDIASUBTYPE_DOLBY_AC3			, "DOLBY_AC3"			},
	{ &MEDIASUBTYPE_DVD_SUBPICTURE		, "DVD_SUBPICTURE"		},
	{ &MEDIASUBTYPE_DVD_LPCM_AUDIO		, "DVD_LPCM_AUDIO"		},
	{ &MEDIASUBTYPE_DVD_NAVIGATION_PCI	, "DVD_NAVIGATION_PCI"	},
	{ &MEDIASUBTYPE_DVD_NAVIGATION_DSI	, "DVD_NAVIGATION_DSI"	},
	{ &MEDIASUBTYPE_DVD_NAVIGATION_PROVIDER		, "DVD_NAVIGATION_PROVIDER"		},
};


//
// CreateInstance
//
// Used by the ActiveMovie base classes to create instances
//
CUnknown *CMFSPropertiesMedia::CreateInstance(LPUNKNOWN lpunk, HRESULT *phr)
{
    CUnknown *punk = new CMFSPropertiesMedia(lpunk, phr);
    if (punk == NULL) {
	*phr = E_OUTOFMEMORY;
    }
    return punk;

} // CreateInstance


//
// Constructor
//
CMFSPropertiesMedia::CMFSPropertiesMedia(LPUNKNOWN pUnk, HRESULT *phr) :
    CBasePropertyPage(NAME("MFSrc filter Property Page - Mediatype"),
                      pUnk,IDD_MEDIATYPE_PROP,IDS_MEDIATYPE_TITLE),
    m_pIPMFSrc(NULL),
	m_pmt(NULL),
    m_bIsInitialized(FALSE)
{
	ASSERT(phr);
} // (Constructor)


//
// OnReceiveMessage
//
// Handles the messages for our property window
//
BOOL CMFSPropertiesMedia::OnReceiveMessage(HWND hwnd,
                                          UINT uMsg,
                                          WPARAM wParam,
                                          LPARAM lParam)
{
    switch (uMsg)
    {
        case WM_COMMAND:
        {
			if(LOWORD(wParam)==IDC_OPEN) 
			{
				OpenMediaTypeFile(hwnd);
			}

            if (m_bIsInitialized)
            {
                m_bDirty = TRUE;
                if (m_pPageSite)
                {
                    m_pPageSite->OnStatusChange(PROPPAGESTATUS_DIRTY);
                }
            }
            return (LRESULT) 1;
        }

    }
    return CBasePropertyPage::OnReceiveMessage(hwnd,uMsg,wParam,lParam);

} // OnReceiveMessage


//
// OnConnect
//
// Called when we connect to a transform filter
//
HRESULT CMFSPropertiesMedia::OnConnect(IUnknown *pUnknown)
{
    ASSERT(m_pIPMFSrc == NULL);

    HRESULT hr = pUnknown->QueryInterface(IID_IIPMFSrc, (void **) &m_pIPMFSrc);
    if (FAILED(hr)) {
        return E_NOINTERFACE;
    }

    ASSERT(m_pIPMFSrc);

    // Get the initial image FX property
	delete m_pmt;
    m_pIPMFSrc->get_MediaType(&m_pmt);
	
    m_bIsInitialized = FALSE ;
    return NOERROR;

} // OnConnect


//
// OnDisconnect
//
// Likewise called when we disconnect from a filter
//
HRESULT CMFSPropertiesMedia::OnDisconnect()
{
    // Release of Interface after setting the appropriate old effect value

    if (m_pIPMFSrc == NULL) {
        return E_UNEXPECTED;
    }

    m_pIPMFSrc->Release();
    m_pIPMFSrc = NULL;
    return NOERROR;

} // OnDisconnect


//
// OnActivate
//
// We are being activated
//
HRESULT CMFSPropertiesMedia::OnActivate()
{
	SetControlValues();

	m_bIsInitialized = TRUE;

    return NOERROR;
} // OnActivate

//
// OnDeactivate
//
// We are being deactivated
//
HRESULT CMFSPropertiesMedia::OnDeactivate(void)
{
    ASSERT(m_pIPMFSrc);
    m_bIsInitialized = FALSE;
    GetControlValues();
    return NOERROR;

} // OnDeactivate


//
// OnApplyChanges
//
// Apply any changes so far made
//
HRESULT CMFSPropertiesMedia::OnApplyChanges()
{
    GetControlValues();
    m_pIPMFSrc->put_MediaType(m_pmt);

    return NOERROR;
} // OnApplyChanges


void CMFSPropertiesMedia::GetControlValues()
{
	HWND	hCB;
	int		index;

    ASSERT(m_pIPMFSrc);

	hCB=GetDlgItem(m_Dlg, IDC_MAJOR);
	index=(int)SendMessage(hCB, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
	m_pmt->SetType(major_types[index].type);

	hCB=GetDlgItem(m_Dlg, IDC_MINOR);
	index=(int)SendMessage(hCB, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
	m_pmt->SetSubtype(minor_types[index].type);
}

void CMFSPropertiesMedia::SetControlValues()
{
	HWND	hCB;
	int		i, index=0;

	hCB=GetDlgItem(m_Dlg, IDC_MAJOR);

	index=0;
	for(i=0; i<(sizeof(major_types)/sizeof(major_types[0])); i++)
	{
		SendMessage(hCB, CB_INSERTSTRING, (WPARAM)(-1), (LPARAM)major_types[i].name);
		if(IsEqualCLSID(*major_types[i].type, *m_pmt->Type()))
			index=i;
	}
	SendMessage(hCB, CB_SETCURSEL, (WPARAM)index, (LPARAM)0);

	hCB=GetDlgItem(m_Dlg, IDC_MINOR);

	index=0;
	for(i=0; i<(sizeof(minor_types)/sizeof(minor_types[0])); i++)
	{
		SendMessage(hCB, CB_INSERTSTRING, (WPARAM)(-1), (LPARAM)minor_types[i].name);
		if(IsEqualCLSID(*minor_types[i].type, *m_pmt->Subtype()))
			index=i;
	}
	SendMessage(hCB, CB_SETCURSEL, (WPARAM)index, (LPARAM)0);
}

static WCHAR	strFile[MAX_PATH]=L"";		// this should be large enough

static OPENFILENAMEW	ofn=
{
	sizeof(OPENFILENAME),		// lStructSize
	NULL,						// hwndOwner
    NULL,						// hInstance
    NULL,						// lpstrFilter 
    NULL,						// lpstrCustomFilter
    0,							// nMaxCustFilter
    0,							// nFilterIndex
    strFile,					// lpstrFile
    sizeof(strFile),			// nMaxFile
    NULL,						// lpstrFileTitle
    0,							// nMaxFileTitle
    NULL,						// lpstrInitialDir
    L"open media type",				// lpstrTitle
    OFN_FILEMUSTEXIST|OFN_LONGNAMES,		// Flags
    0,							// nFileOffset
    0,							// nFileExtension
    NULL,						// lpstrDefExt
    0,							// lCustData
    NULL,						// lpfnHook
    NULL						// lpTemplateName
};

void CMFSPropertiesMedia::OpenMediaTypeFile(HWND hwnd)
{
	ofn.hwndOwner=hwnd;

	if(GetOpenFileNameW(&ofn)==TRUE)
	{
		HRESULT		hr;
		CMediaType	*pmt;

		hr=m_pIPMFSrc->load_MediaType(&pmt, strFile);
	    if(FAILED(hr)) 
		{
			MessageBoxW(hwnd, L"Error loading media type:", strFile, MB_OK|MB_ICONERROR);
			return;
		}

		if(pmt!=NULL)
		{
			if(m_pmt!=NULL)
				delete m_pmt;

			m_pmt=pmt;
			SetControlValues();
		}
	}
}

