//==========================================================================;
//
//  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
//  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
//  PURPOSE.
//
//  Copyright (c) 1996 - 1998  Microsoft Corporation.  All Rights Reserved.
//
//--------------------------------------------------------------------------;

//
//  Define an internal filter that wraps the base CBaseReader stuff
//
#include "../include/asyncio.h"
#include "../include/asyncrdr.h"
#include "mfs_uids.h"
#include "inf_mfs.h"

typedef struct _HFileList
{
	struct _HFileList	*Next;
	
	HANDLE		hFile;
	LONGLONG	offset;		// from beg of file list !!!
	LONGLONG	length;		// the length of this file
} HFileList;

class CMultFileStream : public CAsyncStream
{
public:
    CMultFileStream() :
        m_llPosition(0),
		m_llskip(0),
		m_first(NULL)
    {
    }

	~CMultFileStream()
	{
		FreeFileHandleList();
	}

    /*  Initialization */
	void Init(HFileList *first, LONGLONG skip, DWORD dwKBPerSec = INFINITE)
    {
		HFileList	*fl;

		FreeFileHandleList();

		if(skip<0)
			skip=0;

		m_first=first;
		m_llskip=skip;
        m_llLength = 0;
		fl=m_first;
		while(fl)
		{
			fl->offset=m_llLength;
			m_llLength+=fl->length;

			fl=fl->Next;
		}

        m_dwKBPerSec = dwKBPerSec;
        m_dwTimeStart = timeGetTime();
    }

    HRESULT SetPointer(LONGLONG llPos)
    {
        if (llPos < 0 || llPos > (m_llLength-m_llskip)) {
            return S_FALSE;
        } else {
            m_llPosition = llPos;
            return S_OK;
        }
    }
    HRESULT Read(PBYTE pbBuffer,
                 DWORD dwBytesToRead,
                 BOOL bAlign,
                 LPDWORD pdwBytesRead)
    {
        CAutoLock lck(&m_csLock);
        DWORD dw2Read, dwReadLength;
		LONGLONG	llpos;

		llpos=m_llPosition+m_llskip;

		/*  Wait until the bytes are here! */
        if (llpos + dwBytesToRead > m_llLength) {
            dwBytesToRead = (DWORD)(m_llLength - llpos);
        } 

		(*pdwBytesRead)=0;
        dwReadLength=dwBytesToRead;

		while(dwReadLength>0)
		{
			// first find the correct file to read
			HFileList		*hfl;
			LONGLONG		ll_len;
			LARGE_INTEGER	li;
			DWORD			dwR;

			hfl=m_first;
			while(hfl)
			{
				if(hfl->offset<=llpos && llpos<(hfl->offset+hfl->length))
					break;

				hfl=hfl->Next;
			}

			if(hfl==NULL)
			{
error:
				if((*pdwBytesRead)>0)
					break;
			
				MessageBox(NULL, TEXT("failed"), TEXT("arg"), MB_OK);

				return S_FALSE;
			}

			dw2Read=dwReadLength;
			ll_len=hfl->offset + hfl->length - llpos;
			if(ll_len < dw2Read)
				dw2Read=(DWORD)ll_len;
			
			li.QuadPart=llpos-hfl->offset;
			SetFilePointer (hfl->hFile, li.LowPart, &li.HighPart, FILE_BEGIN);

			dwR=0;
			ReadFile(hfl->hFile, (PVOID)pbBuffer, dw2Read, &dwR, NULL);
			if(dwR==0)
				goto error;

			pbBuffer+=dwR;
			m_llPosition+=dwR;
			llpos+=dwR;
			(*pdwBytesRead)+=dwR;

			dwReadLength-=dwR;
		}
/*
        DWORD dwTime = timeGetTime();
        DWORD dwTimeToArrive =
            ((DWORD)pos + dwBytesToRead) / m_dwKBPerSec;
        if (dwTime - m_dwTimeStart < dwTimeToArrive) {
            Sleep(dwTimeToArrive - dwTime + m_dwTimeStart);
        }
*/
        return S_OK;
    }

    LONGLONG Size(LONGLONG *pSizeAvailable)
    {
 /*       LONGLONG llCurrentAvailable =
            Int32x32To64((timeGetTime() - m_dwTimeStart),m_dwKBPerSec);
        *pSizeAvailable = min(m_llLength, llCurrentAvailable);
		*/

		*pSizeAvailable = (m_llLength-m_llskip);
        return (m_llLength-m_llskip);
    }
    DWORD Alignment()
    {
        return 1;
    }
    void Lock()
    {
        m_csLock.Lock();
    }
    void Unlock()
    {
        m_csLock.Unlock();
    }

	void FreeFileHandleList()
	{
		HFileList	*fl, *nfl;

		fl=m_first;
		while(fl)
		{
			nfl=fl->Next;

			CloseHandle(fl->hFile);
			delete fl;

			fl=nfl;
		}

		m_first=NULL;
	}

private:
    CCritSec       m_csLock;
	HFileList		*m_first;
    LONGLONG       m_llLength;
    LONGLONG       m_llPosition;
    LONGLONG       m_llskip;
    DWORD          m_dwKBPerSec;
    DWORD          m_dwTimeStart;
};

class CMultFileSrcFilter : public CAsyncReader, public IFileSourceFilter, public IIPMFSrc, public ISpecifyPropertyPages
{
public:
    CMultFileSrcFilter(LPUNKNOWN pUnk, HRESULT *phr) :
		CAsyncReader(NAME("Multiple file source reader"), pUnk, &m_Stream, phr),
		m_llskip(0),
		m_first(NULL)
	{
	}


    ~CMultFileSrcFilter()
    {
		FreeFileList();
    }

    static CUnknown *CreateInstance(LPUNKNOWN, HRESULT *);

    DECLARE_IUNKNOWN

    // These implement the custom IIPMFSrc interface
	STDMETHODIMP get_FileList(MFS_FileList **first, LONGLONG *skip);
	STDMETHODIMP put_FileList(MFS_FileList *first, LONGLONG skip);

	STDMETHODIMP load_MediaType(CMediaType **mediaType, WCHAR *filename);
	STDMETHODIMP get_MediaType(CMediaType **mediaType);
	STDMETHODIMP put_MediaType(const CMediaType *mediaType);
	STDMETHODIMP put_MediaType(const AM_MEDIA_TYPE *mediatype);
	STDMETHODIMP put_MediaType(const GUID *majortype, const GUID *minortype);

    // ISpecifyPropertyPages interface
    STDMETHODIMP GetPages(CAUUID *pPages);


    STDMETHODIMP NonDelegatingQueryInterface(REFIID riid, void **ppv);

    /*  IFileSourceFilter methods */
    //  Load a (new) file

    STDMETHODIMP Load(LPCOLESTR lpwszFileName, const AM_MEDIA_TYPE *pmt)
    {
        CheckPointer(lpwszFileName, E_POINTER);

		int				len;
		MFS_FileList	*fl;
	    WCHAR			*lpType;
		CMediaType		mt;

        len=lstrlenW(lpwszFileName);
		fl=(MFS_FileList *)CoTaskMemAlloc(sizeof(MFS_FileList)+len*sizeof(WCHAR));
        if (fl!=NULL) {
			fl->Next=NULL;
		    CopyMemory(fl->Filename, lpwszFileName, (len+1)*sizeof(WCHAR));
        }

		if(fl!=NULL)
		{
			put_FileList(fl, m_llskip);
			CoTaskMemFree(fl);
		}

		if(pmt==NULL)
		{
		    mt.majortype = MEDIATYPE_Stream;

			if(len>=4 && lpwszFileName[len-4]==TEXT('.')) 
			{
				lpType=(WCHAR *)&lpwszFileName[len-3];

				if(lstrcmpW(lpType, L"mpg")==0)
					mt.subtype=MEDIASUBTYPE_MPEG1System;
				else if(lstrcmpW(lpType, L"mpa")==0 || lstrcmpW(lpType, L"mp2")==0 || lstrcmpW(lpType, L"mp3")==0)
					mt.subtype=MEDIASUBTYPE_MPEG1Audio;
				else if(lstrcmpW(lpType, L"mpv")==0)
					mt.subtype=MEDIASUBTYPE_MPEG1Video;
				else if(lstrcmpW(lpType, L"dat")==0)
					mt.subtype=MEDIASUBTYPE_MPEG1VideoCD;
				else if(lstrcmpW(lpType, L"avi")==0)
					mt.subtype=MEDIASUBTYPE_Avi;
				else if(lstrcmpW(lpType, L"mov")==0)
					mt.subtype=MEDIASUBTYPE_QTMovie;
				else if(lstrcmpW(lpType, L"wav")==0)
					mt.subtype=MEDIASUBTYPE_WAVE;
				else if(lstrcmpW(lpType, L"vob")==0)
					mt.subtype=MEDIASUBTYPE_MPEG2_PROGRAM;
			}

			pmt=&mt;
		}
		
		put_MediaType(pmt);


        return S_OK;
    }

    // Modelled on IPersistFile::Load
    // Caller needs to CoTaskMemFree or equivalent.

    STDMETHODIMP GetCurFile(LPOLESTR * ppszFileName, AM_MEDIA_TYPE *pmt)
    {
        CheckPointer(ppszFileName, E_POINTER);
        *ppszFileName=NULL;
        
		if(m_first!=NULL) 
		{
    		DWORD n = sizeof(WCHAR)*(1+lstrlenW(m_first->Filename));

            *ppszFileName = (LPOLESTR) CoTaskMemAlloc( n );
            if (*ppszFileName!=NULL) {
                  CopyMemory(*ppszFileName, m_first->Filename, n);
            }
        }

        if (pmt!=NULL) {
            CopyMediaType(pmt, &m_mt);
        }

        return NOERROR;
    }

private:

	HFileList *createFileHandleList();

	void FreeFileList()
	{
		MFS_FileList	*fl, *nfl;

		fl=m_first;
		while(fl)
		{
			nfl=fl->Next;
			CoTaskMemFree(fl);
			fl=nfl;
		}
		m_first=NULL;
	}

private:
	MFS_FileList	*m_first;		// the first entry in the file list
    LONGLONG       m_llskip;		// how many bytes to skip at the beg of stream

    CMultFileStream m_Stream;
};
